#pragma once

#include <utility>
#include <algorithm>
#include <Qt>
#include <QDebug>
#include <QObject>
#include <QThread>
#include <QMutex>
#include <QElapsedTimer>
#include <QWaitCondition>
#include <bgtu/computer_graphics_lab_work/renderer_api/voxel_drawer.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/sprite.hpp>
#include "owned_qimage.hpp"
#include "_separate_threaded_renderer.hpp"

namespace BGTU::ComputerGraphicsLabWork::QtUtilities {
	template<class sprite_data_t, class renderer_context_t, void (renderer_context_t::*renderer_procedure)(RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::Exporter *, RendererApi::Color, RendererApi::Projection const *, RendererApi::Sprite<sprite_data_t> *const *, std::size_t, sprite_data_t const *)>
	class SeparateThreadedRenderer : public _SeparateThreadedRenderer_Signals {
	private:
		struct data {
		public:
			RendererApi::Sprite<sprite_data_t> *const *sprites;
			std::size_t sprites_count;
			sprite_data_t const *sprite_data;
			std::uint_fast64_t ms_per_frame;
			RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t width, height;
			RendererApi::Projection *projection;
			RendererApi::Color bg;
		};

		class RendererThread : public QThread {

		private:
			SeparateThreadedRenderer *owner;
			QWaitCondition intr;
		public:

			explicit inline RendererThread(SeparateThreadedRenderer *owner) : QThread{owner}, owner{owner}, intr{} {
			};

			void requestInterruption() {
				QThread::requestInterruption();
				this->intr.wakeAll();
			}

		protected:
			void run() final;

		private:
			void sleepms(std::uint_fast64_t ms) {
				QMutex m{};
				m.lock();
				this->intr.wait(&m, ms);
				m.unlock();
			}
		};


		volatile data next_data;
		QMutex sync;
		RendererThread *thread;
		QImage *sentinel_img;
		renderer_context_t *renderer_context;
	public:
		explicit SeparateThreadedRenderer(renderer_context_t *renderer_context, QObject *owner = nullptr);

	public slots:

		void set_frame_size(RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t width, RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t height);

		void set_ms_per_frame(std::uint_fast64_t ms);

		void set_sprites(RendererApi::Sprite<sprite_data_t> *const *sprites, std::size_t sprites_count);

		void set_sprite_data(sprite_data_t const *data);

		void set_sprites_and_data(RendererApi::Sprite<sprite_data_t> *const *sprites, std::size_t sprites_count, sprite_data_t const *data);

		void set_background(RendererApi::Color);

	public:
		~SeparateThreadedRenderer();
	};


	template<class sprite_data_t, class renderer_context_t, void (renderer_context_t::*renderer_procedure)(RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::Exporter *, RendererApi::Color, RendererApi::Projection const *, RendererApi::Sprite<sprite_data_t> *const *, std::size_t, sprite_data_t const *)>
	SeparateThreadedRenderer<sprite_data_t, renderer_context_t, renderer_procedure>::SeparateThreadedRenderer(renderer_context_t *renderer_context, QObject *owner):
			_SeparateThreadedRenderer_Signals{owner},
			sync{},
			renderer_context{renderer_context},
			sentinel_img{new QImage(0, 0, QImage::Format_ARGB32)},
			thread{new RendererThread(this)},
			next_data{
					.sprites = nullptr,
					.sprites_count = 0,
					.sprite_data = nullptr,
					.ms_per_frame = 1000 / 60,
					.width = 0,
					.height = 0,
					.projection = nullptr
			} {
		this->thread->start();
	}

	template<class sprite_data_t, class renderer_context_t, void (renderer_context_t::*renderer_procedure)(RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::Exporter *, RendererApi::Color, RendererApi::Projection const *, RendererApi::Sprite<sprite_data_t> *const *, std::size_t, sprite_data_t const *)>
	void SeparateThreadedRenderer<sprite_data_t, renderer_context_t, renderer_procedure>::set_frame_size(RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t width, RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t height) {
		this->sync.lock();
		const_cast<data *>(&this->next_data)->width = width;
		const_cast<data *>(&this->next_data)->height = height;
		this->sync.unlock();
	}

	template<class sprite_data_t, class renderer_context_t, void (renderer_context_t::*renderer_procedure)(RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::Exporter *, RendererApi::Color, RendererApi::Projection const *, RendererApi::Sprite<sprite_data_t> *const *, std::size_t, sprite_data_t const *)>
	void SeparateThreadedRenderer<sprite_data_t, renderer_context_t, renderer_procedure>::set_ms_per_frame(std::uint_fast64_t ms) {
		this->sync.lock();
		const_cast<data *>(&this->next_data)->ms_per_frame = ms;
		this->sync.unlock();
	}

	template<class sprite_data_t, class renderer_context_t, void (renderer_context_t::*renderer_procedure)(RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::Exporter *, RendererApi::Color, RendererApi::Projection const *, RendererApi::Sprite<sprite_data_t> *const *, std::size_t, sprite_data_t const *)>
	void SeparateThreadedRenderer<sprite_data_t, renderer_context_t, renderer_procedure>::set_sprites(RendererApi::Sprite<sprite_data_t> *const *sprites, std::size_t sprites_count) {
		this->sync.lock();
		const_cast<data *>(&this->next_data)->sprites = sprites;
		const_cast<data *>(&this->next_data)->sprites_count = sprites_count;
		this->sync.unlock();
	}

	template<class sprite_data_t, class renderer_context_t, void (renderer_context_t::*renderer_procedure)(RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::Exporter *, RendererApi::Color, RendererApi::Projection const *, RendererApi::Sprite<sprite_data_t> *const *, std::size_t, sprite_data_t const *)>
	void SeparateThreadedRenderer<sprite_data_t, renderer_context_t, renderer_procedure>::set_sprite_data(sprite_data_t const *sprite_data) {
		this->sync.lock();
		const_cast<data *>(&this->next_data)->sprite_data = sprite_data;
		this->sync.unlock();
	}

	template<class sprite_data_t, class renderer_context_t, void (renderer_context_t::*renderer_procedure)(RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::Exporter *, RendererApi::Color, RendererApi::Projection const *, RendererApi::Sprite<sprite_data_t> *const *, std::size_t, sprite_data_t const *)>
	void SeparateThreadedRenderer<sprite_data_t, renderer_context_t, renderer_procedure>::set_sprites_and_data(RendererApi::Sprite<sprite_data_t> *const *sprites, std::size_t sprites_count, sprite_data_t const *sprite_data) {
		this->sync.lock();
		const_cast<data *>(&this->next_data)->sprites = sprites;
		const_cast<data *>(&this->next_data)->sprites_count = sprites_count;
		const_cast<data *>(&this->next_data)->sprite_data = sprite_data;
		this->sync.unlock();
	}


	template<class sprite_data_t, class renderer_context_t, void (renderer_context_t::*renderer_procedure)(RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::Exporter *, RendererApi::Color, RendererApi::Projection const *, RendererApi::Sprite<sprite_data_t> *const *, std::size_t, sprite_data_t const *)>
	void SeparateThreadedRenderer<sprite_data_t, renderer_context_t, renderer_procedure>::set_background(RendererApi::Color c) {
		this->sync.lock();
		const_cast<data *>(&this->next_data)->bg = c;
		this->sync.unlock();
	}

	template<class sprite_data_t, class renderer_context_t, void (renderer_context_t::*renderer_procedure)(RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::Exporter *, RendererApi::Color, RendererApi::Projection const *, RendererApi::Sprite<sprite_data_t> *const *, std::size_t, sprite_data_t const *)>
	void SeparateThreadedRenderer<sprite_data_t, renderer_context_t, renderer_procedure>::RendererThread::run() {
		try {
			QElapsedTimer timer;
			while (!this->isInterruptionRequested()) {
				timer.start();
				this->owner->sync.lock();
				data cached = *const_cast<data *>(&this->owner->next_data);
				this->owner->sync.unlock();

				OwnedQImage img{cached.width, cached.height};
				(this->owner->renderer_context->*renderer_procedure)(
						cached.width, cached.height,
						&img,
						cached.bg,
						cached.projection,
						cached.sprites, cached.sprites_count,
						cached.sprite_data
				);
				emit this->owner->frame_rendered(img);
				this->sleepms(std::max<std::uint_fast64_t>(cached.ms_per_frame - static_cast<std::uint_fast64_t>(timer.elapsed()), 0));
			}
		} catch (std::exception const &e) {
			qCritical() << typeid(e).name() << ": " << e.what() << "\n" << "Renderer thread died" << "\n";
		}
	}


	template<class sprite_data_t, class renderer_context_t, void (renderer_context_t::*renderer_procedure)(RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::Exporter *, RendererApi::Color, RendererApi::Projection const *, RendererApi::Sprite<sprite_data_t> *const *, std::size_t, sprite_data_t const *)>
	SeparateThreadedRenderer<sprite_data_t, renderer_context_t, renderer_procedure>::~SeparateThreadedRenderer() {
		this->thread->requestInterruption();
		this->thread->wait();
	}
}